/*
 * led-power - main.c
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <adc.h>
#include <jiffies.h>
#include <watchdog.h>
#include <limits.h>
#include <io.h>
#include <stdio.h>

/*
 * BD244 temperature control set points.
 *
 * The transistor is turned off transistor if its temperature is above
 * T_HIGH over ambient temperature; it is turned on if its temperature
 * drops below T_LOW.
 *
 * High temperatures limit the maximal power dissipation due to
 * junction-to-case resistance. For 79 C it is about 36.7 W,
 * due to delay in control we limit the temperature to 77 C.
 * We assume that the ambient temperature is 45 C.
 */
#define T_HIGH		32000
#define T_LOW		12000

/*
 * Maximal DC input voltage for normal operation. Under high temperatures
 * we are power-limited to 33.8 W (see T_HIGH) due to BD244 junction-to-case
 * resistance. It's hard limit for us, we do not depend on internal junction
 * thermal capacitance. Assume that the maximal current is 1.2 A, and we
 * drop 0.3 V on diode and 0.564 V on resistor.
 */
#define VIN_MAX		30220

/*
 * special short-circuit protection
 */
#define T_SHORT		17000
#define V_SHORT		1000

#define V_MED		200

/*
 * the initial temperature
 */
#define T_INIT		T_HIGH

/*
 * Parameters for current estimation. Based on linear approximation
 * of the voltage/current characteristic at 24V input voltage
 * from SPICE model.
 */
#define V_D	300	/* mV, voltage drop on B540C */
#define I_MAX	1059	/* mA, saturation current */
#define V_MAX	1230	/* mV, minimal voltage for saturation */

#define I_SHIFT	4

/*
 * estimate_ic - estimate the input current
 * @vd: voltage difference, in millivolts (mV).
 *
 * This function returns the estimated current in milliamps (mA).
 *
 * We use simple linear model with two saturations. Below V_D
 * the current is equal to zero, because the voltage is too low
 * for series diode. Above V_MAX the current is constant.
 */
static unsigned int estimate_ic(unsigned int vd)
{
	unsigned int ic = 0;

	if (vd <= V_D)
		return 0;
	if (vd >= V_MAX)
		return I_MAX;
	vd -= V_D;
	ic = ((I_MAX << I_SHIFT) / (V_MAX - V_D)) * vd >> I_SHIFT;
	return ic;
}

/*
 * estimate_power - estimate the power on BD244
 * @vd: voltage difference, in millivolts (mV).
 * @ic: collector current, in milliamps (mA).
 *
 * This function returns the estimated power in milliwats (mW).
 */
static unsigned int estimate_power(unsigned int vd, unsigned int ic)
{
	unsigned int v_other;
	unsigned char v;
	unsigned char i;

	if (vd >= VIN_MAX)
		return 40000;

	/*
	 * voltage on other elements, 0.25V as a safe value on diode,
	 * and the voltage on 0.47 resistor (0.47 approximated
	 * with 15/32 = 0.4687).
	 */
	v_other = 250 + (15 * ic >> 5);
	if (v_other >= vd)
		v_other = vd;
	vd -= v_other;

	/* assume at least 256 mV on BD244 */
	if (vd < 256)
		vd = 256;

	v = vd >> 7;
	i = (ic + 7) >> 3;

	return v * i;
}

#define T_MAX	(65536 - 1024)

static unsigned int q1_thermal_model(unsigned int p)
{
	static unsigned int temp = T_INIT;
	static unsigned int temp_e;
	static unsigned char p_e;

	/*
	 * BD244 temperature control
	 *
	 * For low voltages the BD244 is maximal current-limited,
	 *
	 * BC244 thermal model
	 *
	 *         R P
	 * T = ----------
	 *      RC s + 1
	 *
	 * T - temperature over ambient temperature (45 C assumed),
	 *     in millikelvins (mK)
	 * C - thermal capacitance, 0.54 J/K, from Fairchild AN-7516
	 * R - thermal resistivity, assumed 42 K/W based on
	 *	simplified PCB FEM thermal model (with 50% safety margin)
	 * 80, from Fairchild AN-7516.
	 * P - transistor power, assumed 40W, maximal C * P = 21.6
	 *
	 * After discretization using backward differences with HZ=61.035
	 * frequency:
	 *
	 *         R P
	 * T = ----------------
	 *      R C (1 - q) HZ + 1
	 *
	 * where q = z^-1.
	 *
	 * (R C HZ + 1) T - R C HZ q T = R P
	 * (R C HZ + 1) T = R C HZ q T + R P
	 * (R C HZ + 1) T = (R C HZ + 1) q T - q T + R P
	 *
	 *               q T           R P
	 * T = q T - ----------- + ------------
	 *            R C HZ + 1    R C HZ + 1
	 *
	 * R C HZ + 1 = 1400, approximated with 1365
	 * R / (R C HZ + 1) = 0.0300, approximated with 1/32 = 0.03125
	 *
	 * To increase precision we use simple error shaper, the previous
	 * quantization error is added to input in the next sample.
	 */
	temp_e += temp;
	temp -= 3 * (temp_e >> 12);
	temp_e &= 0xfff;

	p += p_e;
	temp += p >> 5;
	p_e = p & 0x1f;
	if (temp > T_MAX)
		temp = T_MAX;
	return temp;
}

void wdt_init(void)
{
	watchdog_init();
}

int main(void)
{
	unsigned char t;
	unsigned int vin;
	unsigned int vout;
	unsigned int ic;
	unsigned int p;
	unsigned char enabled = 0;
	unsigned int temp;

	wdt_init();
	io_init();
	adc_init();
	jiffies_init();

	wdt_reset();

	t = get_jiffies();
	for (;;) {
		wdt_reset();

		/*
		 * We implement discrete-time control, the I/O phase
		 * should be as short as possible.
		 */
		vin = vin_get();
		vout = vout_get();
		switch_set(enabled);

		/*
		 * estimate the collector current and power dissipated
		 * on transistor
		 */
		ic = 0;
		p = 0;
		if (enabled) {
			if (vin > vout) {
				unsigned int vd = vin - vout;

				ic = estimate_ic(vd);
				p = estimate_power(vd, ic);
			}

			/* additional 41 milliwats for base current */
			p += 41;
		}

		/*
		 * estimate the transistor temperature
		 */
		temp = q1_thermal_model(p);

		/*
		 * simple controller
		 */
		if (vin >= VIN_MAX)
			enabled = 0;
		else if (temp > T_HIGH)
			enabled = 0;
		else if (temp > T_SHORT && vout < V_SHORT)
			enabled = 0;
		else if (temp < T_LOW)
			enabled = 1;

		logger(vin, vout, ic, p, temp, enabled);
		io_xmit(!enabled);

		t++;
		sleep_to(t);
	}
}
